/***************************************************************
 *
 *  ParticleSimulation.cpp
 *
 *  (c) 2007 Jonas Spillmann, Matthias Teschner
 *      Computer Graphics, University of Freiburg
 *
 **************************************************************/

#pragma warning(disable:4244)
#pragma warning(disable:4018)

#include "definitions.h"
#include "textoutput.h"
#include "precisetimer.h"
#include "particle.h"

DclVector getWorld();



// Integration step h
const DclFloat  TIMESTEP = 0.005;

// Elasticity of collision with walls
const DclFloat  WORLD_ELASTICITY = 1.0;



//////////// GLOBALS /////////////////
std::vector<Particle> vectorParticles;
int frame = 0;

//////////////////////////////////////

void  visualizeParticles();

//------------------------------------------------------------------------------------------------------------------------------------

// initialize the particle properties
void  initializeSimulation()
{
  DclVector world = getWorld();

  srand(0);

  for (int i = 0; i < 100; i ++)
  {
    Particle  particle;

    // TODO: Initialize physical properties of particles
    // You can use the function rand() to retrieve a random number between 0 and RAND_MAX,
    // and you can use srand(0) to reset the random number generator
    // Also look at the structure definition in particle.h!


    // put the particle into the vector
    vectorParticles.push_back(particle);
  }
}

//------------------------------------------------------------------------------------------------------------------------------------

// add external forces to the sum of forces acting on the particle
void  addForces(Particle& particle)
{
  // TODO: Compute gravitational force of particle
  //particle.force += ... // gravitational force
}

//------------------------------------------------------------------------------------------------------------------------------------

// perform one time-integration step
void  integrate(Particle& particle)
{
  switch (particle.integration)
  {
    case Euler:
      // TODO: Implement Euler scheme
      break;

    case EulerCromer:
      // TODO: Implement Euler-Cromer scheme
      break;

	case Verlet:
		// TODO: Implement Verlet scheme
		break;

    default:
      break;
  }
}

//------------------------------------------------------------------------------------------------------------------------------------

// collision handling with the world
void  worldCollision(Particle& particle)
{
  DclVector world = getWorld();

  world *= 0.5;

  if (particle.position.x < -world.x || particle.position.x > world.x)
  {
    particle.velocity.x *= -WORLD_ELASTICITY;
    particle.position.x = SIGN(particle.position.x) * world.x;
  }

  if (particle.position.y < -world.y || particle.position.y > world.y)
  {
    particle.velocity.y *= -WORLD_ELASTICITY;
    particle.position.y = SIGN(particle.position.y) * world.y;
  }

  if (particle.position.z < -world.z || particle.position.z > world.z)
  {
    particle.velocity.z *= -WORLD_ELASTICITY;
    particle.position.z = SIGN(particle.position.z) * world.z;
  }
}

//------------------------------------------------------------------------------------------------------------------------------------

// compute and output the total energies in the scene
void  computeEnergies()
{
  DclFloat  e_kin = 0, e_pot = 0;
  DclVector world = getWorld();

  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    Particle& particle = vectorParticles[part];

    // TODO: Compute kinetic and potential energy
    //e_kin += ...
    //e_pot += ...
  }

  Textout::printvar(6, 0, "Kinetic energy %f, Potential energy %f, Sum %f", e_kin, e_pot, e_kin + e_pot);
}

//------------------------------------------------------------------------------------------------------------------------------------

// perform one simulation step
void  simulationStep()
{
  // reset the vector of external forces
  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    vectorParticles[part].force.clear();
  }

  // compute external forces on the particles
  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    addForces(vectorParticles[part]);
  }

  // compute collision impulses
  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    worldCollision(vectorParticles[part]);
  }

  // perform one integration step
  for (int part = 0; part < vectorParticles.size(); part ++)
  {
    integrate(vectorParticles[part]);
  }

  // compute and output energies
  computeEnergies();

  // draw particles
  visualizeParticles();

  // measure time for frame
  double  dt = PreciseTimer::stop("SimulationStep");
  PreciseTimer::start("SimulationStep");

  // output statistics
  Textout::printvar(5, 0, "Frame %d, Frames/second %f, Num. particles %d", frame, 1000.0 / dt, vectorParticles.size());
  frame ++;
}


